#include "guide.h"
#include "text_over.h"
#include "orange-channels.h"
#include "thread.h"
#include "utils.h"
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>

#define gprintf(...) cgprintf(255, 255, 255, __VA_ARGS__)

extern channels channels_list;

static character *screen;
static int maxlines;
static int column;
static int line;
static int start_line;

static int space_in_front_allowed;

static void gputchar(int c, int r, int g, int b)
{
  if (c == ' ' && (column == 0 || column == 80))
    if (!space_in_front_allowed) return;
  if (column == 80) {
    column = 0;
    line++;
  }
  while (line >= maxlines) {
    maxlines += 20;
    screen = realloc(screen, maxlines * 80 * sizeof(character));
    if (screen == NULL) abort();
    memset(screen + (maxlines-20) * 80, 0, 20 * 80 * sizeof(character));
  }
  if (c == '\n') {
    line++;
    column = 0;
    return;
  }
  screen[line*80+column].c = c;
  screen[line*80+column].r = r;
  screen[line*80+column].g = g;
  screen[line*80+column].b = b;
  column++;
}

static int wordlen(char *s)
{
  int l = 0;
  int v;
  while (*s && !isspace(*s)) {
    s += get_utf8(s, &v);
    l++;
  }
  return l;
}

static void cgprintf(int r, int g, int b, char *s, ...)
#ifdef __GNUC__
   __attribute__ ((format (printf, 4, 5)))
#endif
;

void cgprintf(int r, int g, int b, char *s, ...)
{
  int i;
  static char *buf = NULL;
  static int len = 0;
  int last_is_space = 0;
  int tlen;
  va_list ap;
  while (1) {
    va_start(ap, s);
    tlen = vsnprintf(buf, len, s, ap);
    va_end(ap);
    if (tlen < len) break;
    len += 1024;
    buf = realloc(buf, len * sizeof(character));
    if (buf == NULL) abort();
  }
  for (i = 0; buf[i];) {
    int v;
    if (!isspace(buf[i]) && last_is_space) {
      int l = wordlen(buf+i);
      if (l < 60 && column != 80 && l + column > 80)
        gputchar('\n', r, g, b);
    }
    i += get_utf8(buf+i, &v);
    gputchar(v, r, g, b);
    last_is_space = isspace(v);
  }
}

static void put_screen(void)
{
  character l[81];
  character *s;
  int i, j;

  ofull();
  for (i = 0; i < 22 && i+start_line < line; i++) {
    s = screen + (start_line + i) * 80;
    for (j = 0; j < 80 && s[j].c; j++) l[j] = s[j];
    l[j].c = 0;
    oputline(i, l);
  }
}

static int guide_channel;

void show_guide(int chan)
{
  channel *c = &channels_list.c[chan];
  int i;

  lock_channels_list();

  guide_channel = chan;
  memset(screen, 0, maxlines * 80 * sizeof(character));

  column = 0;
  line = 0;
  start_line = 0;

  cgprintf(128, 128, 255,
           "ESC/g: quit guide     Up/Down/PageUp/PageDown: scroll text\n");
  cgprintf(128, 128, 255,
           "Left/Right: show guide for previous/next channel\n\n");
  cgprintf(255, 128, 128, "%s\n\n", c->name);
  for (i = 0; i < c->psize; i++) {
    if (c->p[i].title == NULL) continue;
    cgprintf(128, 255, 128,
             "[%2.2dh%2.2d - %2.2dh%2.2d] %s\n",
             c->p[i].hstart,
             c->p[i].mstart,
             c->p[i].hend,
             c->p[i].mend,
             c->p[i].title);
    if (c->p[i].genre)
      gprintf("%s\n", c->p[i].genre);
    if (c->p[i].summary)
      cgprintf(180, 180, 255, "\n%s\n\n", c->p[i].summary);
  }
  cgprintf(128, 128, 255, "[END]\n");

  unlock_channels_list();

  put_screen();
}

void next_guide(int delta)
{
  while (delta) {
    if (delta < 0) guide_channel--; else guide_channel++;
    if (guide_channel < 0) guide_channel = channels_list.size - 1;
    if (guide_channel >= channels_list.size) guide_channel = 0;
    delta -= delta>0 ? 1 : -1;
  }
  show_guide(guide_channel);
}

void show_full_guide(void)
{
  int chan;
  channel *c;
  int i;

  lock_channels_list();

  space_in_front_allowed = 1;

  chan = 0;

  column = 0;
  line = 0;
  start_line = 0;

  memset(screen, 0, maxlines * 80 * sizeof(character));

  cgprintf(128, 128, 255, "ESC/g: quit full guide     "
           "Up/Down/PageUp/PageDown: scroll text\n\n");

  for (chan = 0; chan < channels_list.size; chan++) {
    c = &channels_list.c[chan];

    cgprintf(255, 128, 128, "%s\n\n", c->name);
    for (i = 0; i < c->psize; i++) {
      if (c->p[i].title == NULL) continue;
      cgprintf(128, 255, 128,
               "  [%2.2dh%2.2d - %2.2dh%2.2d] %s\n",
               c->p[i].hstart,
               c->p[i].mstart,
               c->p[i].hend,
               c->p[i].mend,
               c->p[i].title);
    }
    gprintf("\n");
  }

  cgprintf(128, 128, 255, "[END]\n");

  space_in_front_allowed = 0;

  unlock_channels_list();

  put_screen();
}

void scroll_guide(int delta)
{
  start_line += delta;
  if (start_line < 0) start_line = 0;
  if (start_line >= line) start_line = line-1;
  put_screen();
}

static volatile int reload_in_progress;

static void *reload_guide_thread(void *_)
{
  extern int no_drm;
  extern int no_no_signal;
  channels c;
  int i, j;

  oclear();
  oprintf("reloading guide...\n");

  memset(&c, 0, sizeof(c));
  get_channels(&c, no_drm, no_no_signal);

  lock_channels_list();

  for (i = 0; i < c.size; i++)
    for (j = 0; j < channels_list.size; j++)
      if (!strcmp(c.c[i].id, channels_list.c[j].id)) {
        /* exchange old programs with new ones */
        program *p = c.c[i].p;
        int psize = c.c[i].psize;
        c.c[i].p = channels_list.c[i].p;
        c.c[i].psize = channels_list.c[i].psize;
        channels_list.c[i].p = p;
        channels_list.c[i].psize = psize;
      }

  unlock_channels_list();

  free_channels(&c);

  oclear();
  oprintf("guide reloaded\n");
  odelay(500);

  reload_in_progress = 0;
  return NULL;
}

void reload_guide(void)
{
  lock_channels_list();
  if (reload_in_progress) { unlock_channels_list(); return; }
  reload_in_progress = 1;
  unlock_channels_list();
  new_thread(reload_guide_thread, NULL);
}

void show_help(void)
{
  space_in_front_allowed = 1;
  column = 0;
  line = 0;
  start_line = 0;

  memset(screen, 0, maxlines * 80 * sizeof(character));
  cgprintf(128, 128, 255, "ESC/h: quit help     "
           "Up/Down/PageUp/PageDown: scroll text\n\n");

  cgprintf(180,180,255, "Commands:\n");
  cgprintf(128,255,128, "    a");
  cgprintf(180,180,255, "   aspect ratio\n");
  cgprintf(128,255,128, "    e");
  cgprintf(180,180,255, "   equalizer on/off\n");
  cgprintf(128,255,128, "    E");
  cgprintf(180,180,255, "   equalizer's controls\n");
  cgprintf(128,255,128, "    f");
  cgprintf(180,180,255, "   fullscreen on/off\n");
  cgprintf(128,255,128, "    F");
  cgprintf(180,180,255, "   set window size to current video size\n");
  cgprintf(128,255,128, "    g");
  cgprintf(180,180,255, "   channel's guide\n");
  cgprintf(128,255,128, "    G");
  cgprintf(180,180,255, "   global guide\n");
  cgprintf(128,255,128, "    h");
  cgprintf(180,180,255, "   help\n");
  cgprintf(128,255,128, "    m");
  cgprintf(180,180,255, "   mute/unmute\n");
  cgprintf(128,255,128, "    M");
  cgprintf(180,180,255, "   reset volume\n");
  cgprintf(128,255,128, "    q");
  cgprintf(180,180,255, "   quit\n");
  cgprintf(128,255,128, "    s");
  cgprintf(180,180,255, "   audio/video sync: audio back in time\n");
  cgprintf(128,255,128, "    S");
  cgprintf(180,180,255, "   audio/video sync: audio advance in time\n");
  cgprintf(128,255,128, "    v");
  cgprintf(180,180,255, "   volume down\n");
  cgprintf(128,255,128, "    V");
  cgprintf(180,180,255, "   volume up\n");
  cgprintf(128,255,128, "    DOWN arrow");
  cgprintf(180,180,255, "      volume down\n");
  cgprintf(128,255,128, "    UP arrow");
  cgprintf(180,180,255, "        volume up\n");
  cgprintf(128,255,128, "    LEFT arrow");
  cgprintf(180,180,255, "      previous channel\n");
  cgprintf(128,255,128, "    RIGHT arrow");
  cgprintf(180,180,255, "     next channel\n");
  cgprintf(128,255,128, "    mouse's wheel");
  cgprintf(180,180,255, "   volume up/down\n");
  cgprintf(128,255,128, "    1/2/3/4/5/6/7/8/9/0");
  cgprintf(180,180,255, "   set channel 1...10\n");
  cgprintf(128,255,128, "    CTRL+g");
  cgprintf(180,180,255, "   reload guide\n");

  cgprintf(180,180,255, "\nIn the equalizer's controls:\n");
  cgprintf(128,255,128, "    ESC");
  cgprintf(180,180,255, "   close the controls' window\n");
  cgprintf(128,255,128, "    MOUSE");
  cgprintf(180,180,255, "   use it to set the equalizer to your taste\n");

  gprintf("\n");
  cgprintf(128, 128, 255, "[END]\n");

  space_in_front_allowed = 0;

  put_screen();
}
