#include "h264.h"
#include "video_buffer.h"
#include "utils.h"
#include "mp4parser.h"
#include <setjmp.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <libavcodec/avcodec.h>
#include <libswscale/swscale.h>
#include "logger.h"

typedef struct {
  AVPacket p;
  AVCodec *codec;
  AVCodecContext *c;
  AVFrame *frame;
  char *header;
  int header_size;
} ffmpeg_data;

static struct SwsContext *ctx;
static unsigned char *image;
static int img_width, img_height;

static uint64_t ffmpeg_eat_data(ffmpeg_data *f,
        uint64_t timestamp, uint64_t timescale, int *frame_count)
{
  int len, got_frame;
#if 0
{
  static FILE *oo = NULL;
  if (oo == NULL) oo = fopen("out.video", "w"); if (!oo) abort();
  fwrite(f->p.data, f->p.size, 1, oo);
  fflush(oo);
}
#endif
//fwrite(f->p.data, f->p.size, 1, stdout);
  len = avcodec_decode_video2(f->c, f->frame, &got_frame, &f->p);
  if (len < 0) { logprintf("Error while decoding frame (err %d 0x%x '%s')\n", len, len, av_err2str(len)); /*abort();*/ return timestamp; }
  if (len != f->p.size) abort();
  if (got_frame) {
    int h;
    int image_linesize;
    if (frame_count != NULL) *frame_count = *frame_count + 1;
    if (f->c->width != img_width || f->c->height != img_height) {
      if (ctx != NULL) sws_freeContext(ctx);
      free(image);
      ctx = sws_getContext(f->c->width, f->c->height, f->c->pix_fmt,
                           f->c->width, f->c->height, AV_PIX_FMT_RGBA,
                           0, NULL, NULL, NULL);
      if (ctx == NULL) abort();
      image = malloc(f->c->width*f->c->height*4);
      if (image == NULL) abort();
      img_width = f->c->width;
      img_height = f->c->height;
    }
    logprintf("got frame width %d height %d format "
              "%d AV_PIX_FMT_YUVJ420P=%d!\n",
        f->c->width, f->c->height, f->frame->format, AV_PIX_FMT_YUVJ420P);
    logprintf("pts %"PRIu64" pkt_pts %"PRIu64" coded_picture_number "
              "%u display_picture_number %u key_frame %d pict_type %d\n",
        f->frame->pts, f->frame->pkt_pts, f->frame->coded_picture_number,
        f->frame->display_picture_number, f->frame->key_frame,
        f->frame->pict_type);
    image_linesize = 4 * f->c->width;
    h = sws_scale(ctx, (const uint8_t**)f->frame->data,
                  f->frame->linesize, 0, f->c->height,
                  &image, &image_linesize);
    if (h != f->c->height) abort();
    put_image((char*)image, f->c->width, f->c->height, timestamp, timescale);
    /* TODO: get FPS from stream, we hardcode 25 */
    extern volatile double fps;
    timestamp += timescale / fps; //25;
    if (its_over) return timestamp;
    //void display_image(char *, int, int);
    //display_image((char *)image, f->c->width, f->c->height);
  }
  return timestamp;
}

static char *buffer;
static int size;
static int pos;
static jmp_buf j;

static uint32_t getu32(void)
{
  uint32_t r;
  if (size - pos < 4) longjmp(j, 1);
  r = ((unsigned char)buffer[pos] << 24) |
      ((unsigned char)buffer[pos+1] << 16) |
      ((unsigned char)buffer[pos+2] << 8) |
      ((unsigned char)buffer[pos+3]);
  pos += 4;
  return r;
}

static uint64_t getu64(void)
{
  uint64_t r;
  if (size - pos < 8) longjmp(j, 1);
  r = ((uint64_t)((unsigned char)buffer[pos]) << (32+24)) |
      ((uint64_t)((unsigned char)buffer[pos+1]) << (32+16)) |
      ((uint64_t)((unsigned char)buffer[pos+2]) << (32+8)) |
      ((uint64_t)((unsigned char)buffer[pos+3]) << 32) |
      ((uint64_t)((unsigned char)buffer[pos+4]) << 24) |
      ((uint64_t)((unsigned char)buffer[pos+5]) << 16) |
      ((uint64_t)((unsigned char)buffer[pos+6]) << 8) |
      ((uint64_t)((unsigned char)buffer[pos+7]));
  pos += 8;
  return r;
}

uint64_t dump_mdat(video_data *v, uint64_t box_start, uint64_t box_size,
        uint64_t timestamp, int *frame_count)
{
  ffmpeg_data *f;
  int saved_size = size;
  uint32_t mdat_item_size;

  f = v->ffmpeg;

  size = box_start + box_size;

  while (pos < size) {
    mdat_item_size = getu32();
    buffer[pos-4] = 0;
    buffer[pos-3] = 0;
    buffer[pos-2] = 0;
    buffer[pos-1] = 1;
    if (pos + mdat_item_size > size) longjmp(j, 1);
//    fprintf(stdout, "%c%c%c%c", 0, 0, 0, 1);
//    fwrite(buffer + pos, mdat_item_size, 1, stdout);
//    fflush(stdout);

    f->p.data = (unsigned char *)buffer + pos - 4;
    f->p.size = mdat_item_size + 4;
    timestamp = ffmpeg_eat_data(f, timestamp, v->timescale, frame_count);
    if (its_over) return timestamp;

    pos += mdat_item_size;
  }

  size = saved_size;

  return timestamp;
}

int dump_video(video_data *v, char *b, int _size)
{
  int frame_count = 0;
  uint64_t timestamp = get_fragment_start_time(b, _size);
  buffer = b;
  size = _size;
  pos = 0;

  if (setjmp(j)) { logprintf("error with video data\n"); return frame_count; }

  while (pos < size) {
    int dapos = pos;
    uint64_t boxsize = getu32();
    uint32_t boxtype = getu32();
    if (boxsize == 1) boxsize = getu64();
    if (boxsize == 0) boxsize = size - dapos;
    logprintf("box %c%c%c%c size %"PRIu64"\n",
        (boxtype >> 24) & 255,
        (boxtype >> 16) & 255,
        (boxtype >> 8) & 255,
        (boxtype) & 255, boxsize);
    if (boxtype == 0x6d646174)
      timestamp = dump_mdat(v, dapos, boxsize, timestamp, &frame_count);
    if (its_over) return frame_count;

    pos = dapos + boxsize;
//FILE *f = fopen("ZOB", "w"); fwrite(buffer, _size, 1, f); fclose(f); exit(1);
  }

  return frame_count;
}

void dump_video_header(video_data *v, char *p)
{
  ffmpeg_data *f = v->ffmpeg;
  char *p2;
  char *pdec;
  int len;
  logprintf("dump_video_header full p %s\n", p);
  pdec = malloc(strlen(p)); if (pdec == NULL) abort();
  p2 = strstr(p+2, "00000001"); if (p2) *p2 = 0;
again:
  logprintf("dump_video_header p %s\n", p);
  memset(pdec, 0, strlen(p));
  len = 0;
  while (*p) {
    int a = (unsigned char)p[0];
    int b = (unsigned char)p[1];
    p += 2;
    if (a >= '0' && a <= '9') a = a - '0';
    else if (a >= 'A' && a <= 'F') a = 10 + a - 'A';
    else if (a >= 'a' && a <= 'f') a = 10 + a - 'a';
    else abort();
    if (b >= '0' && b <= '9') b = b - '0';
    else if (b >= 'A' && b <= 'F') b = 10 + b - 'A';
    else if (b >= 'a' && b <= 'f') b = 10 + b - 'a';
    else abort();
    pdec[len] = (a<<4) | b;
    len++;
  }
  logprintf("len %d\n", len);
  f->p.size = len;
  f->p.data = (unsigned char *)pdec;
  ffmpeg_eat_data(f, 0, 0, NULL);
  if (p2 != NULL) { *p2 = '0'; p = p2; p2 = NULL; goto again; }
  free(pdec);
  //f->header = pdec;
  //f->header_size = len;
}

#include <libavformat/avformat.h>

void global_init_video(void)
{
  avcodec_register_all();
}

void init_video_data(video_data *v, uint64_t timescale)
{
  ffmpeg_data *f;
  f = malloc(sizeof(ffmpeg_data)); if (f == NULL) abort();
  v->timescale = timescale;
  v->ffmpeg = f;
  av_init_packet(&f->p);
  f->codec = avcodec_find_decoder(AV_CODEC_ID_H264);
  if (f->codec == NULL) { logprintf("no H264 codec\n"); abort(); }
  f->c = avcodec_alloc_context3(f->codec);
  if (f->c == NULL)
    { logprintf("avcodec_alloc_context3 error\n"); abort(); }
  if (f->codec->capabilities & CODEC_CAP_TRUNCATED)
    f->c->flags |= CODEC_FLAG_TRUNCATED;
  f->c->flags2 |= CODEC_FLAG2_CHUNKS;
  if (avcodec_open2(f->c, f->codec, NULL) < 0)
    { logprintf("Could not open codec\n"); abort(); }
  f->frame = av_frame_alloc();
  if (!f->frame)
    { logprintf("Could not allocate video frame\n"); exit(1); }
}

void free_video_data(video_data *v)
{
  ffmpeg_data *f = v->ffmpeg;

  free(image);          image = NULL;
  sws_freeContext(ctx); ctx = NULL;
  img_width = 0;
  img_height = 0;

  av_frame_free(&f->frame);
  avcodec_close(f->c);
  avcodec_free_context(&f->c);

  free(v->ffmpeg); v->ffmpeg = NULL;
}
