#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <inttypes.h>
#include <time.h>
#include "orange-channels.h"
#include "download.h"
#include "jsmn.h"
#include "xml.h"

static int get_element(char *s, jsmntok_t *t, int dad, char *elem)
{
  int i;
  int cur = dad+1;
  if (t[dad].type != JSMN_OBJECT) return -1;

  for (i = 0; i < t[dad].size; i++) {
    if (t[cur].type != JSMN_STRING) return -1;
    if (!strncmp(elem, s+t[cur].start, t[cur].end-t[cur].start))
      return cur+1;
    cur++;
    if (i != t[dad].size - 1) while (t[cur].parent != dad) cur++;
  }

  return -1;
}

static int get_array(jsmntok_t *t, int dad, int index)
{
  int i;
  int cur = dad+1;
  for (i = 0; i < index; i++) {
    cur++;
    while (t[cur].parent != dad) cur++;
  }
  return cur;
}

static char *get_string(char *s, jsmntok_t *t, int elem)
{
  char *r;

  if (elem == -1) return NULL;

  if (t[elem].type != JSMN_STRING &&
      t[elem].type != JSMN_PRIMITIVE) return NULL;

  r = malloc(t[elem].end-t[elem].start+1); if (r == NULL) abort();
  memcpy(r, s+t[elem].start, t[elem].end-t[elem].start);
  r[t[elem].end-t[elem].start] = 0;

  return r;
}

static int get_bool(char *s, jsmntok_t *t, int elem)
{
  if (elem == -1) return -1;
  if (t[elem].type != JSMN_PRIMITIVE) return -1;
  if (!strncmp(s+t[elem].start, "true", 4)) return 1;
  if (!strncmp(s+t[elem].start, "false", 5)) return 0;
  return -1;
}

static void get_programs(channel *ch, char *s, jsmntok_t *t, int item)
{
  int i;
  int p = get_element(s, t, item, "programs");
  if (p == -1) return;
  if (t[p].type != JSMN_ARRAY) return;
  ch->psize = t[p].size;
  ch->p = malloc(ch->psize * sizeof(program));
  if (ch->p == NULL) abort();
  memset(ch->p, 0, ch->psize * sizeof(program));
  for (i = 0; i < ch->psize; i++) {
    uint64_t start, end;
    uint64_t dd;
    time_t dt;
    struct tm *dtm;
    int hstart, mstart, hend, mend;
    int prog = get_array(t, p, i);
    int title, date, duration, genre, summary, additional_data;
    if (t[prog].type != JSMN_OBJECT) return;
    title = get_element(s, t, prog, "title");
    date = get_element(s, t, prog, "diffusionDate");
    duration = get_element(s, t, prog, "duration");
    genre = get_element(s, t, prog, "genre");
    additional_data = get_element(s, t, prog, "additionnalData");
    summary = get_element(s, t, additional_data, "longSummary");
    if (title == -1 || date == -1 || duration == -1 || genre == -1 ||
        additional_data == -1 || summary == -1) abort();
    //dd = atoll(s+t[date].start) / 1000;
    sscanf(s+t[date].start, "%"SCNu64, &dd);
    dd /= 1000;
    start = dd;
    dt = dd;
    dtm = localtime(&dt);
    hstart = dtm->tm_hour;
    mstart = dtm->tm_min;
    //dd = atoll(s+t[duration].start);
    sscanf(s+t[duration].start, "%"SCNu64, &dd);
    dt += dd * 60;
    end = dt;
    dtm = localtime(&dt);
    hend = dtm->tm_hour;
    mend = dtm->tm_min;
    ch->p[i].start = start;
    ch->p[i].hstart = hstart;
    ch->p[i].mstart = mstart;
    ch->p[i].end = end;
    ch->p[i].hend = hend;
    ch->p[i].mend = mend;
    ch->p[i].title = malloc(t[title].end-t[title].start+1);
    if (ch->p[i].title == NULL) abort();
    memcpy(ch->p[i].title, s+t[title].start, t[title].end-t[title].start);
    ch->p[i].title[t[title].end-t[title].start] = 0;
    ch->p[i].genre = malloc(t[genre].end-t[genre].start+1);
    if (ch->p[i].genre == NULL) abort();
    memcpy(ch->p[i].genre, s+t[genre].start, t[genre].end-t[genre].start);
    ch->p[i].genre[t[genre].end-t[genre].start] = 0;
    ch->p[i].summary = malloc(t[summary].end-t[summary].start+1);
    if (ch->p[i].summary == NULL) abort();
    memcpy(ch->p[i].summary, s+t[summary].start,
           t[summary].end-t[summary].start);
    ch->p[i].summary[t[summary].end-t[summary].start] = 0;
  }
}

static int has_signal(channel *c)
{
  int ret = 1;
  dbuf b;
  char *manifestURL = get_channel_url(c);
  b.size = b.maxsize = 0; b.data = NULL; b.error = 0;
  if (manifestURL == NULL) { ret = 0; goto end; }
  download_unstoppable(&b, manifestURL);
  if (b.size == 0) { ret = 0; goto end; }
  if (try_parse_xml(b.data, b.size)) ret = 0;
end:
  free(manifestURL);
  free(b.data);
  return ret;
}

/* c is supposed emptied when entering this function */
static void parse_service_plan(channels *c, char *b, int bsize,
    int no_drm, int no_no_signal)
{
  jsmn_parser p;
  jsmntok_t *tokens = NULL;
  jsmnerr_t r;
  int response;
  int i;
  int nb_channels;

  /* first parsing to get # of tokens */
  jsmn_init(&p);
  r = jsmn_parse(&p, b, bsize, NULL, 0); if (r < 0) goto error;
  tokens = malloc(r * sizeof(jsmntok_t)); if (tokens == NULL) abort();
  /* second "real" parsing */
  jsmn_init(&p);
  r = jsmn_parse(&p, b, bsize, tokens, r); if (r < 0) goto error;
  if (tokens[0].type != JSMN_OBJECT) goto error;

  response = get_element(b, tokens, 0, "response");
  if (response == -1 || tokens[response].type != JSMN_ARRAY) goto error;
  nb_channels = tokens[response].size;

  c->c = malloc(nb_channels * sizeof(*c->c)); if (c->c == NULL) abort();
  memset(c->c, 0, nb_channels * sizeof(*c->c));
  c->size = 0; //nb_channels;

  for (i = 0; i < nb_channels; i++) {
    channel *ch =  &c->c[c->size];
    int item = get_array(tokens, response, i);
    int name = get_element(b, tokens, item, "name");
    int id   = get_element(b, tokens, item, "channelId");
    int drm  = get_element(b, tokens, item, "drmized");
    ch->drm = get_bool(b, tokens, drm); if (ch->drm == -1) goto error;
    ch->name = get_string(b, tokens, name); if (ch->name == NULL) goto error;
    ch->id = get_string(b, tokens, id); if (ch->id == NULL) goto error;
    if ((no_drm && ch->drm) || (no_no_signal && !has_signal(ch)))
      { printf("INFO: skip channel '%s' (%d/%d)\n", ch->name,i+1,nb_channels);
        free(ch->name); free(ch->id); continue; }
    printf("INFO: OK   channel '%s' (%d/%d)\n", ch->name, i+1, nb_channels);
    get_programs(ch, b, tokens, item);
#if 0
    dump_programs(s, tokens, item);
#endif
    c->size++;
  }

  free(tokens);
  return;

error:
  free_channels(c);
  free(tokens);
}

char *service_plan = SERVICE_PLAN;

void get_channels(channels *c, int no_drm, int no_no_signal)
{
  dbuf b;

  memset(&b, 0, sizeof(b));
  download_unstoppable(&b, service_plan);

  parse_service_plan(c, b.data, b.size, no_drm, no_no_signal);

  free(b.data);
}

void free_channels(channels *c)
{
  int i, j;
  for (i = 0; i < c->size; i++) {
    program *p = c->c[i].p;
    free(c->c[i].name);
    free(c->c[i].id);
    for (j = 0; j < c->c[i].psize; j++) {
      free(p[j].title);
      free(p[j].genre);
      free(p[j].summary);
    }
  }
  free(c->c);
  c->c = NULL;
  c->size = 0;
}

static char *parse_channel_url(char *b, int bsize)
{
  jsmn_parser p;
  jsmntok_t *tokens = NULL;
  jsmnerr_t r;
  int response;
  int url;
  char *ret;

  /* first parsing to get # of tokens */
  jsmn_init(&p);
  r = jsmn_parse(&p, b, bsize, NULL, 0); if (r < 0) goto error;
  tokens = malloc(r * sizeof(jsmntok_t)); if (tokens == NULL) abort();
  /* second "real" parsing */
  jsmn_init(&p);
  r = jsmn_parse(&p, b, bsize, tokens, r); if (r < 0) goto error;
  if (tokens[0].type != JSMN_OBJECT) goto error;

  response = get_element(b, tokens, 0, "response");
  if (response == -1 || tokens[response].type != JSMN_OBJECT) goto error;
  url = get_element(b, tokens, response, "url");
  ret = get_string(b, tokens, url);

  free(tokens);

  return ret;

error:
  free(tokens);
  return NULL;
}

char *get_channel_url(channel *c)
{
  char *r;
  dbuf b;
  char *url = malloc(strlen(CHANNEL_INFO_URL) + strlen(c->id));
  if (url == NULL) abort();
  sprintf(url, CHANNEL_INFO_URL, c->id);
  memset(&b, 0, sizeof(b));
  download(&b, url);
  free(url);
  r = parse_channel_url(b.data, b.size);
  free(b.data);
  return r;
}
